--[[
Specialization for using fuel with the Kaercher HPW

Author: 	Marhu
Co-Author: 	Ifko[nator]
Date: 		20.02.2016
Version: 	1.0

History:	V1.0 @ 20.02.2016 - intial release
			----------------------------------
]]

PressureWasherFuelUsage = {};

function PressureWasherFuelUsage.prerequisitesPresent(specializations)
	return SpecializationUtil.hasSpecialization(PressureWasher, specializations);
end;

function PressureWasherFuelUsage:load(xmlFile)
	self.setIsFuelFilling = SpecializationUtil.callSpecializationsFunction("setIsFuelFilling");
    self.setFuelFillLevel = SpecializationUtil.callSpecializationsFunction("setFuelFillLevel");
    self.addFuelFillTrigger = PressureWasherFuelUsage.addFuelFillTrigger;
	self.removeFuelFillTrigger = PressureWasherFuelUsage.removeFuelFillTrigger;
	self.IsFillActivatable = PressureWasherFuelUsage.IsFillActivatable;
	
	self.fuelCapacity = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.fuelCapacity"), 50);
    self.fuelUsage = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.fuelUsage"), 1) / 36000;
	self.motorizedFillActivatable = MotorizedRefuelActivatable:new(self);
	
	self.motorizedFillActivatable.getIsActivatable = function(...)
		return self:IsFillActivatable(...) 
	end;
    
	self.fuelFillTriggers = {};
    self.isFuelFilling = false;
    self.fuelFillLitersPerSecond = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.fuelFillLitersPerSecond"), 5);
    self:setFuelFillLevel(0);
	self.sentFuelFillLevel = self.fuelFillLevel;

	if self.isClient then
		local linkNode = Utils.indexToObject(self.components, Utils.getNoNil(getXMLString(xmlFile, "vehicle.refuelDieselSound#linkNode"),  "0>"))
		self.sampleRefuelDiesel = Utils.loadSample(xmlFile, {}, "vehicle.refuelDieselSound", "$data/maps/sounds/refuel.wav", self.baseDirectory, linkNode);
	end;

	self.FuelUsageDirtyFlag = self:getNextDirtyFlag();
end;

function PressureWasherFuelUsage:getSaveAttributesAndNodes(nodeIdent)
	local attributes = 'fuelFillLevel="' .. self.fuelFillLevel .. '"';
    
	return attributes, nil;
end

function PressureWasherFuelUsage:loadFromAttributesAndNodes(xmlFile, key, resetVehicles)
	local fuelFillLevel = getXMLFloat(xmlFile, key .. "#fuelFillLevel");
	local fillLevel = getXMLFloat(xmlFile, key .. "#fillLevel");
    
	if fuelFillLevel ~= nil then
        self:setFuelFillLevel(fuelFillLevel);
	end;
	
	if fillLevel ~= nil and fillLevel > self.capacity then
        self:setFillLevel(self.capacity, Fillable.FILLTYPE_WATER);
	end;
	
	return BaseMission.VEHICLE_LOAD_OK;
end

function PressureWasherFuelUsage:writeUpdateStream(streamId, connection, dirtyMask)
	if not connection.isServer then
        if streamWriteBool(streamId, bitAND(dirtyMask, self.FuelUsageDirtyFlag) ~= 0) then
            local percent = Utils.clamp(self.fuelFillLevel / self.fuelCapacity, 0, 1);
            
			streamWriteUIntN(streamId, math.floor(percent * 32767), 15);
        end;
	end;
end;

function PressureWasherFuelUsage:readUpdateStream(streamId, timestamp, connection)
	if connection.isServer then
        if streamReadBool(streamId) then
            local fuelFillLevel = streamReadUIntN(streamId, 15) / 32767 * self.fuelCapacity;
          
			self:setFuelFillLevel(fuelFillLevel);
        end;
	end;
end;

function PressureWasherFuelUsage:delete()
end;

function PressureWasherFuelUsage:update(dt)
	local messageInRange;
	
	if self.currentPlayer ~= nil then
		messageInRange = self:getIsPlayerInRange(self.actionRadius + 1, self.currentPlayer);
	end;
	
	if self:getIsActiveForInput() or (messageInRange and self.currentPlayer == g_currentMission.player) then
		if self.fuelFillLevel ~= nil then
			g_currentMission:addExtraPrintText(string.gsub(g_i18n:getText("CurrentFuelFillLevel"), "currentFuelFillLevel", math.floor(self.fuelFillLevel)));
		end;
	end;
	
	if self.fuelFillLevel <= 0 then
		if self.currentPlayer == g_currentMission.player then	
			g_currentMission:showBlinkingWarning(g_i18n:getText("HighPressureWasherFuelIsEmpty"), 4000);
		end;
	end;
	
	if self.isServer then
		if self.isTurnedOn then
			local fuelUsedFac = 8;
			
			if self.doWashing then 
				fuelUsedFac = fuelUsedFac * 2;
			end;
		
			local fuelUsed = fuelUsedFac * (self.fuelUsage * dt);
			
			self:setFuelFillLevel(self.fuelFillLevel - fuelUsed);
			
			if fuelUsed > 0 then 
				g_currentMission.missionStats:updateStats("fuelUsage", fuelUsed); 
			end;
			
			if self.fuelFillLevel <= 0 then
				self:setIsTurnedOn(false, nil);
			end;
		end;
	end;
end;

function PressureWasherFuelUsage:updateTick(dt)	
	if self.isServer then
		if self.isFuelFilling then
            local delta = 0;
            if self.fuelFillTrigger ~= nil then
                delta = self.fuelFillLitersPerSecond * dt * 0.001;
                delta = self.fuelFillTrigger:fillFuel(self, delta);
            end;
			
            if delta <= 0.001 then
                self:setIsFuelFilling(false);
            end;
        end;
		
		if math.abs(self.fuelFillLevel-self.sentFuelFillLevel) > 0.001 then
            self:raiseDirtyFlags(self.FuelUsageDirtyFlag);
            self.sentFuelFillLevel = self.fuelFillLevel;
		end;
	end;
end;

function PressureWasherFuelUsage:setFuelFillLevel(newFillLevel)
    self.fuelFillLevel = math.max(math.min(newFillLevel, self.fuelCapacity), 0);
end;

function PressureWasherFuelUsage:setIsFuelFilling(isFilling, noEventSend)
    if isFilling ~= self.isFuelFilling then
        if noEventSend == nil or noEventSend == false then
            if g_server ~= nil then
                g_server:broadcastEvent(SteerableToggleRefuelEvent:new(self, isFilling), nil, nil, self);
            else
                g_client:getServerConnection():sendEvent(SteerableToggleRefuelEvent:new(self, isFilling));
            end;
        end;
		
        self.isFuelFilling = isFilling;
        
		if isFilling then
            self.fuelFillTrigger = nil;
            for i=1, #self.fuelFillTriggers do
                local trigger = self.fuelFillTriggers[i];
                if trigger:getIsActivatable(self) then
                    self.fuelFillTrigger = trigger;
                    break;
                end;
            end;
        end;
		
        if self.isClient and self.sampleRefuelDiesel ~= nil then
            if isFilling then
                Utils.play3DSample(self.sampleRefuelDiesel);
            else
                Utils.stop3DSample(self.sampleRefuelDiesel);
            end;
        end;
    end;
end;

function PressureWasherFuelUsage:addFuelFillTrigger(trigger)
	if #self.fuelFillTriggers == 0 then
        g_currentMission:addActivatableObject(self.motorizedFillActivatable);
    end;
	
    table.insert(self.fuelFillTriggers, trigger);
end;

function PressureWasherFuelUsage:removeFuelFillTrigger(trigger)
    for i = 1, #self.fuelFillTriggers do
        if self.fuelFillTriggers[i] == trigger then
            table.remove(self.fuelFillTriggers, i);
            
			break;
        end;
    end;
    
	if #self.fuelFillTriggers == 0 or trigger == self.fuelFillTrigger then
        if self.isServer then
            self:setIsFuelFilling(false);
        end;
       
	   if #self.fuelFillTriggers == 0 then
            g_currentMission:removeActivatableObject(self.motorizedFillActivatable);
        end;
    end;
end;

function PressureWasherFuelUsage:IsFillActivatable(me)
	 if self:getIsActiveForInput(false, true) and self.fuelFillLevel-1 < self.fuelCapacity then
        for i = 1, #self.fuelFillTriggers do
            local trigger = self.fuelFillTriggers[i];
           
		   if trigger:getIsActivatable(self) then
                me:updateActivateText();
                
				return true;
            end;
        end;
    end;
	
	return false;
end;

function PressureWasherFuelUsage:readStream(streamId, connection)
	local isFuelFilling = streamReadBool(streamId);
	local newFuelFillLevel = streamReadFloat32(streamId);
		
	self:setIsFuelFilling(isFuelFilling, true);
	self:setFuelFillLevel(newFuelFillLevel);
end;

function PressureWasherFuelUsage:writeStream(streamId, connection)
	streamWriteBool(streamId, self.isFuelFilling);
	streamWriteFloat32(streamId, self.fuelFillLevel);
end;


FuelUsageSetDieselEvent = {}

FuelUsageSetDieselEvent_mt = Class(FuelUsageSetDieselEvent, Event)
InitEventClass(FuelUsageSetDieselEvent, "FuelUsageSetDieselEvent")

function FuelUsageSetDieselEvent:emptyNew()
	local self = Event:new(FuelUsageSetDieselEvent_mt);
	return self;
end;

function FuelUsageSetDieselEvent:new(object, isTurnedOn)
	local self = FuelUsageSetDieselEvent:emptyNew();
	self.object = object;
	self.isTurnedOn = isTurnedOn;
	return self;
end;

function FuelUsageSetDieselEvent:readStream(streamId, connection)
	local id = streamReadInt32(streamId);
	self.isTurnedOn = streamReadBool(streamId);
	self.object = networkGetObject(id);
	self:run(connection);
end;

function FuelUsageSetDieselEvent:writeStream(streamId, connection)
	streamWriteInt32(streamId, networkGetObjectId(self.object));
	streamWriteBool(streamId, self.isTurnedOn);
end;

function FuelUsageSetDieselEvent:run(connection)
	if not connection:getIsServer() then
		g_server:broadcastEvent(self, false, connection, self.object);
	end;

	if self.object ~= nil then
		self.object:setisTurnedOn(self.isTurnedOn, true);
	end;
end;

function FuelUsageSetDieselEvent.sendEvent(vehicle, isTurnedOn, noEventSend)
	if isTurnedOn ~= vehicle.isTurnedOn and (noEventSend == nil or noEventSend == false) then
		if g_server ~= nil then
			g_server:broadcastEvent(FuelUsageSetDieselEvent:new(vehicle, isTurnedOn), nil, nil, vehicle);
		else
			g_client:getServerConnection():sendEvent(FuelUsageSetDieselEvent:new(vehicle, isTurnedOn));
		end;
	end;
end;

--## unused functions

function PressureWasherFuelUsage:mouseEvent(posX, posY, isDown, isUp, button)end;
function PressureWasherFuelUsage:keyEvent(unicode, sym, modifier, isDown)end;
function PressureWasherFuelUsage:draw()end;